;;; rrse --- Ruby Reference|Refe Searcher for Emacsen

;; Copyright (C) 2006 Ohbayashi Ippei

;; Version: 0.1
;; URL: http://www.kmc.gr.jp/~ohai/rrse.html

;; This software is distributed under GNU general Public Licence version 2.
;; 
;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.



(defvar rrse-program-name
  (if (featurep 'meadow)
      "ruby"
    "rrse-refe-listdescs"))

(defvar rrse-program-args
  (if (featurep 'meadow)
      '("-S" "rrse-refe-listdescs")
    nil))

(defvar rrse-refe-program-name
  (if (featurep 'meadow)
      "ruby"
    "refe"))

(defvar rrse-refe-program-args
  (if (featurep 'meadow)
      '("-S" "refe")
    nil))

(defvar rrse-refe-buffer-name "*rrse-refe*")

(defvar rrse-forward-table nil)
(defvar rrse-rev-table nil)

(defvar rrse-enable t)

(defvar rrse-idle-delay 0.5)

(defun rrse-id-char-p (c)
  (or (and (<= ?a c) (<= c ?z))
      (and (<= ?A c) (<= c ?Z))
      (and (<= ?0 c) (<= c ?9))
      (= ?_ c)))

(defun rrse-call-process (buf &rest args)
  (let ((coding-system-for-read 'euc-japan))
    (apply 'call-process rrse-program-name nil buf nil
           (append rrse-program-args args))))

(defun rrse-refe-call-process (buf &rest args)
    (let ((coding-system-for-read 'euc-japan))
    (apply 'call-process rrse-refe-program-name nil buf nil
           (append rrse-program-args args))))

(defun rrse-each-line (fun)
  (goto-char (point-min))
  (while (not (eobp))
    (funcall fun (thing-at-point 'line))
    (forward-line)))

(defun rrse-chomp-string (str)
  (if (eq (aref str (- (length str) 1)) ?\n)
      (substring str 0 (- (length str) 1))
    str))
  
(defun rrse-make-table ()
  (with-temp-buffer
    (rrse-call-process t)
    (let ((table (make-vector (count-lines (point-min) (point-max)) nil)))
      (rrse-each-line (lambda (line)
                        (aset table (count-lines (point-min) (point))
                              (split-string line "[\t\n]"))))
      table)))
    
(defun rrse-make-rev-table ()
  (with-temp-buffer
    (rrse-call-process t "-r")
    (let ((table (make-vector (count-lines (point-min) (point-max)) nil)))
      (rrse-each-line (lambda (line)
                        (aset table (count-lines (point-min) (point))
                              (split-string line "[\t\n]"))))
      table)))

(defun rrse-prepare-table ()
  (if (null rrse-forward-table)
      (setq rrse-forward-table (rrse-make-table)
            rrse-rev-table (rrse-make-rev-table))))

(defun rrse-bsearch (table word begin end)
  "binary search vector table"
  (if (or (= (+ begin 1) end) (= begin end))
      (cond
       ((string= word (car (aref table begin))) (cdr (aref table begin)))
       ((string= word (car (aref table end))) (cdr (aref table end)))
       (t nil))
    (let ((center (/ (+ begin end) 2)))
      (if (string< (car (aref table center)) word)
          (rrse-bsearch table word (+ center 1) end)
        (rrse-bsearch table word begin center)))))

(defun rrse-search-word (word)
  (rrse-bsearch rrse-rev-table word 0 (- (length rrse-rev-table) 1)))

(defun rrse-get-current-word ()
  (save-excursion
    (buffer-substring-no-properties
     (progn (skip-syntax-backward "w_") (point))
     (progn (skip-syntax-forward "w_") (point)))))

(defun rrse-find-dot ()
  (save-excursion
    (char-before (progn (skip-syntax-backward "w_") (point)))))

(defun rrse-filter (fun list)
  (cond
   ((null list) nil)
   ((funcall fun (car list)) (cons (car list) (rrse-filter fun (cdr list))))
   (t (rrse-filter fun (cdr list)))))

(defun rrse-valid-spec-p (word spec dot reciever)
  (or (and (eq dot ?.)
           (or (and (not (string= word "new"))
                    (string-match (concat "#" (regexp-quote word)) spec))
               (and reciever
                    (not (string= "" reciever))
                    (string-match (concat (regexp-quote reciever)
                                          "."
                                          (regexp-quote word))
                                  spec))))
      (and (string-match "^\\(Object#\\|Kernel#\\|Module#\\)" spec)
           (not (string-match word "class\\|initialize")))))
      
(defun rrse-format-descs (descs)
  (mapconcat (lambda (desc) desc) descs " "))

(defun rrse-get-specs (word dot reciever)
  (rrse-filter (lambda (spec)
                 (rrse-valid-spec-p word spec dot reciever))
               (rrse-search-word word)))


(defun rrse-specs-to-descs (specs)
  (let ((make-desc (lambda (desc) (concat spec desc)))
        (pick-desc (lambda (spec)
                     (mapcar make-desc
                             (rrse-bsearch rrse-forward-table
                                           spec
                                           0
                                           (- (length rrse-forward-table) 1))))))
    (apply 'append (mapcar pick-desc specs))))

(defun rrse-find-reciever ()
  (if (not (eq (rrse-find-dot) ?.))
      ""
    (save-excursion
      (let ((end nil))
        (skip-syntax-backward "w_")
        (backward-char)
        (setq end (point))
        (skip-syntax-backward "w_")
        (buffer-substring-no-properties (point) end)))))

(defun rrse-show-minibuffer-help ()
  (let* ((word (rrse-get-current-word))
         (dot (rrse-find-dot))
         (reciever (rrse-find-reciever))
         (specs (rrse-get-specs word dot reciever))
         (descs (rrse-specs-to-descs specs)))
    (if descs
        (message "%s" (rrse-format-descs descs)))))
                 

(defun rrse-print-current-symbol-info ()
  (interactive)
  (and rrse-enable
       (eq major-mode 'ruby-mode)
       (not executing-kbd-macro)
       (not (eq (selected-window) (minibuffer-window)))
       (sit-for rrse-idle-delay)
       (rrse-show-minibuffer-help)))

(defun rrse-show-help (method)
  (let ((buf (or (get-buffer rrse-refe-buffer-name)
                 (generate-new-buffer rrse-refe-buffer-name))))
    (set-buffer buf)
    (setq buffer-read-only nil)
    (erase-buffer)
    (rrse-refe-call-process buf method)
    (goto-char (point-min))
    (set-buffer-modified-p nil)
    (setq buffer-read-only t)
    (display-buffer buf)))
    
(defun rrse-help-completing-read (methods)
  (completing-read "Method: " (mapcar 'list methods)))

(defun rrse-help ()
  (interactive)
  (let ((methods (rrse-get-specs (rrse-get-current-word)
                                 (rrse-find-dot)
                                 (rrse-find-reciever))))
    (cond
     ((null methods) nil) 
     ((null (cdr methods)) (rrse-show-help (car methods)))
     (t (rrse-show-help (rrse-help-completing-read methods))))))
      
(defun rrse-setup ()
  (interactive)
  (add-hook 'ruby-mode-hook
            '(lambda ()
               (rrse-prepare-table)
               (add-hook 'post-command-hook 'rrse-print-current-symbol-info))))
